//& *** (c) 2006-2011 The HPGCC3 Team ***
//& Claudio Lapilli
//& Ingo Blank
//&
//& This file is licensed under the terms and conditions of the
//& HPGCC3 license that is included with the source distribution.
//& *** (c) 2006-2011 The HPGCC3 Team ***

// $Header: fsgetnextentry.c,r42 2006-02-18 04:23:47 ingo $


#include "fsyspriv.h"



// FILLS IN A FS_FILE STRUCTURE WITH THE NEXT DIRECTORY ENTRY
// STARTS AT CurrentOffset

int FSGetNextEntry(FS_FILE *entry,FS_FILE *dir)
{
int order,diroffset;
int nentries;
char checksum;
char buffer[32],*morebuff,*ptr;
FS_VOLUME *fs;

if(!FSystem.Init) return FS_ERROR;
if(!dir || !entry) return FS_ERROR;

fs=FSystem.Volumes[dir->Volume];

if(!fs) return FS_ERROR;

while(FSReadLL(buffer,32,dir,fs)==32)
{
if(buffer[0]==0) return FS_EOF;
if(buffer[0]==0xe5) continue;	// DELETED ENTRY, USE NEXT ENTRY
if( (buffer[11]&FSATTR_LONGMASK) == FSATTR_LONGNAME) {

	// TREAT AS LONG FILENAME
	if(!(buffer[0]&0X40)) continue;		// ORPHAN ENTRY, SKIP
	diroffset=dir->CurrentOffset-32;
	// FOUND LAST ENTRY OF A NAME
	nentries=buffer[0]&0x3f;
	morebuff=(char *)malloc(32*nentries);
	if(morebuff==NULL) return FS_ERROR;
	if(FSReadLL(morebuff,32*nentries,dir,fs)!=32*nentries) { free(morebuff); return FS_ERROR; }
	
	// VERIFY THAT ENTRIES ARE VALID
	ptr=morebuff;
	for(order=nentries-1;(order!=0)&& ((*ptr&0x3f)==order) ;--order,ptr+=32)
	{
		if(ptr[13]!=buffer[13]) break;		// VERIFY CHECKSUM
	}
	if(order) {
		// ENTRIES ARE ORPHANS, DISCARD AND CONTINUE SEARCHING
		FSSeek(dir,-32*(order+1),SEEK_CUR);		// REWIND TO NEXT UNKNOWN ENTRY
		free(morebuff);
		continue;
	}
	// VERIFY THAT SHORT ENTRY FOLLOWS LONG NAME
	
	if( ((ptr[11]&FSATTR_LONGMASK) == FSATTR_LONGNAME) || (*ptr==0) || (*ptr==0xe5)) {

	// VALID SHORT ENTRY NOT FOUND
	free(morebuff);
	if(*ptr==0) return FS_EOF;
	if(*ptr!=0xe5) FSSeek(dir,-32,SEEK_CUR);		// REWIND LAST ENTRY
	continue;
	}
	
	// CALCULATE CHECKSUM
	checksum=0;
	for(order=0;order<11;++order,++ptr)
	{
	checksum= (((checksum<<7)&0x80) | ((checksum>>1)&0x7f)) + *ptr;
	}
	
	if(checksum!=buffer[13]) {
	// FAILED CHECKSUM, SKIP ORPHANS AND CONTINUE
	free(morebuff);
	FSSeek(dir,-32,SEEK_CUR);		// REWIND LAST ENTRY
	continue;
	}
	// VALID ENTRY FOUND, FILL STRUCTURE AND RETURN
	entry->Name=(char *)malloc(nentries*13+1);
	if(entry->Name==NULL) {
	free(morebuff);
	return FS_ERROR;
	}
	
	ptr-=11;
	// REPACK LONG NAME
	for(order=1;order<nentries;++order)
	{
	FSPackName(entry->Name+13*(order-1),ptr-(order<<5));
	}
	FSPackName(entry->Name+13*(order-1),buffer);
	entry->Name[13*nentries]=0;		// FORCE NULL-TERMINATED STRING

	memcpy(buffer,ptr,32);
	free(morebuff);
	
	}
	
	else {
	// IT'S A SHORT NAME ENTRY
	diroffset=dir->CurrentOffset-32;
	nentries=0;
	entry->Name=(char *)malloc(13);
	if(entry->Name==NULL) return FS_ERROR;
	FSPackShortName(entry->Name,buffer);
	}
	
	if(FSystem.CaseMode==FSCASE_SENSHP) FSStripSemi(entry->Name);
	// NOW FILL THE COMMON FIELDS
	entry->Volume=fs->VolNumber;
	entry->Mode=0;
	entry->Attr=buffer[11];
	entry->NTRes=buffer[12];
	entry->CrtTmTenth=buffer[13];
	entry->LastAccDate=ReadInt16((char *)buffer+18);
	entry->CreatTimeDate=ReadInt32((char *)buffer+14);
	entry->WriteTimeDate=ReadInt32((char *)buffer+22);
	entry->FirstCluster=buffer[26]+(buffer[27]<<8)+(buffer[20]<<16)+(buffer[21]<<24);
	entry->FileSize=ReadInt32((char *)buffer+28);
	entry->CurrentOffset=0;
	entry->DirEntryOffset=diroffset;
	entry->DirEntryNum=nentries+1;
	entry->Dir=dir;
	memset((void *)&(entry->Chain),0,sizeof(FS_FRAGMENT)+sizeof(FS_BUFFER));
	return FS_OK;
}
	
if(dir->CurrentOffset>=dir->FileSize) return FS_EOF;
else return FS_ERROR;
}

