//& *** (c) 2006-2011 The HPGCC3 Team ***
//& Claudio Lapilli
//& Ingo Blank
//&
//& This file is licensed under the terms and conditions of the
//& HPGCC3 license that is included with the source distribution.
//& *** (c) 2006-2011 The HPGCC3 Team ***

#include <hpgcc3.h>
extern "C" {
#include <ggl.h>
}

#include "memblock.h"
#include "gstring.h"
#include "gui.h"
#include "resources.h"



// CLASS gForm IMPLEMENTATION

gForm::gForm(gControl& _Parent,const char *_Caption,gFormStyles _Style) : gControl(_Parent)
{
	int w,h;
	StateFlags=0;
	SoftMenu=NULL;
	Focus=NULL;
	
	w=Parent->sizex;
	h=Parent->sizey;
	
	Caption=(char *)_Caption;
	StyleFlags=(gFormStyles)(_Style | S_FULLSCREEN);
	scrollh=scrollv=FALSE;
	ncx=0;
	ncy=0;
	ncwidth=w;
	ncheight=h;

	// CALCULATE CLIENT AREA HEIGHT BASED ON STYLES
	
	if(StyleFlags & S_FORMBORDER) { posx+=BORDERWIDTH; posy+=BORDERWIDTH; h-=2*BORDERWIDTH; w-=2*BORDERWIDTH; }
	if(StyleFlags & (S_TOPCAPTION|S_BOTTOMCAPTION) ) {
		h-=CAPTIONHEIGHT;
		if(StyleFlags & S_TOPCAPTION) posy+=CAPTIONHEIGHT;
	}
	if(StyleFlags & S_FORCEHSCROLL) { h-=SCROLLBARWIDTH; scrollh=TRUE; }
	if(StyleFlags & S_FORCEVSCROLL) { scrollv=TRUE; w-=SCROLLBARWIDTH; }
	
	// CALCULATE CLIENT AREA WIDTH BASED ON STYLES
	
	if(w<0) w=1;
	if(h<0) h=1;
	
	viewx=viewy=0;
	vieww=w;
	viewh=h;

	if(!Resize(w,h)) InsufMemoryError();

	ClearClipRegion();

}
gForm::gForm(gControl& _Parent,const char *_Caption,gFormStyles _Style,int clientwidth,int clientheight)
: gControl(_Parent)
{
	int w,h;

	StateFlags=0;
	SoftMenu=NULL;
	Focus=NULL;

	
	w=Parent->sizex;
	h=Parent->sizey;
	
	Caption=(char *)_Caption;
	StyleFlags=(gFormStyles)(_Style | S_FULLSCREEN);

	ncx=0;
	ncy=0;
	ncwidth=w;
	ncheight=h;

	// CALCULATE CLIENT AREA HEIGHT BASED ON STYLES
	
	if(StyleFlags & S_FORMBORDER) { posx+=BORDERWIDTH; posy+=BORDERWIDTH; h-=2*BORDERWIDTH; w-=2*BORDERWIDTH; }
	if(StyleFlags & (S_TOPCAPTION|S_BOTTOMCAPTION) ) {
		h-=CAPTIONHEIGHT;
		if(StyleFlags & S_TOPCAPTION) posy+=CAPTIONHEIGHT;
	}
	if(StyleFlags & (S_SINGLESOFTMENU)) h-=SOFTMENUHEIGHT;
	else if(StyleFlags & (S_EXSOFTMENU)) h-=3*SOFTMENUHEIGHT;
	
	scrollh=(StyleFlags&S_FORCEHSCROLL)? 1:0;
	scrollv=(StyleFlags&S_FORCEVSCROLL)? 1:0;
	
	if(clientwidth>w) scrollh=1;
	if(clientheight>h) scrollv=1;

	if(scrollh) {
		h-=SCROLLBARWIDTH;
		if(!scrollv) {
			if(clientheight>h) scrollv=1;
		}
	}
	if(scrollv) {
		w-=SCROLLBARWIDTH;
		if(!scrollh) {
			if(clientwidth>w) {
				h-=SCROLLBARWIDTH;
				scrollh=1;
			}
		}
	}
	
	vieww=w;
	viewh=h;
	viewx=viewy=0;
	
	if(!Resize(clientwidth,clientheight)) InsufMemoryError();

	ClearClipRegion();

}


gForm::gForm(gControl& Parent,const char *_Caption,gFormStyles _Style,int x,int y,int width,int height)
: gControl(Parent)
{
	int w,h;

	StateFlags=0;
	SoftMenu=NULL;
	Focus=NULL;

	
	w=width;
	h=height;
	
	Caption=(char *)_Caption;
	StyleFlags=_Style;

	posx=x;
	posy=y;
	ncx=x;
	ncy=y;
	ncwidth=w;
	ncheight=h;

	// CALCULATE CLIENT AREA HEIGHT BASED ON STYLES
	
	if(StyleFlags & S_FORMBORDER) { posx+=BORDERWIDTH; posy+=BORDERWIDTH; h-=2*BORDERWIDTH; w-=2*BORDERWIDTH; }
	if(StyleFlags & (S_TOPCAPTION|S_BOTTOMCAPTION) ) {
		h-=CAPTIONHEIGHT;
		if(StyleFlags & S_TOPCAPTION) posy+=CAPTIONHEIGHT;
	}
	
	scrollh=(StyleFlags&S_FORCEHSCROLL)? 1:0;
	scrollv=(StyleFlags&S_FORCEVSCROLL)? 1:0;
	
	if(scrollh) {
		h-=SCROLLBARWIDTH;
	}
	if(scrollv) {
		w-=SCROLLBARWIDTH;
	}
	
	if(w<0) w=1;
	if(h<0) h=1;
	vieww=w;
	viewh=h;
	viewx=viewy=0;
	
	
	if(!Resize(w,h)) InsufMemoryError();

	ClearClipRegion();

}

gForm::gForm(gControl& Parent,const char *_Caption,gFormStyles _Style,int x,int y,int width,int height,int clientwidth,int clientheight)
	: gControl(Parent)
	{
		int w,h;

		StateFlags=0;
		SoftMenu=NULL;
		Focus=NULL;
	
		w=width;
		h=height;
		
		Caption=(char *)_Caption;
		StyleFlags=_Style;

		ncx=x;
		ncy=y;
		posx=x;
		posy=y;
		ncwidth=w;
		ncheight=h;

		// CALCULATE CLIENT AREA HEIGHT BASED ON STYLES
		
		if(StyleFlags & S_FORMBORDER) { posx+=BORDERWIDTH; posy+=BORDERWIDTH; h-=2*BORDERWIDTH; w-=2*BORDERWIDTH; }
		if(StyleFlags & (S_TOPCAPTION|S_BOTTOMCAPTION) ) {
			h-=CAPTIONHEIGHT;
			if(StyleFlags & S_TOPCAPTION) posy+=CAPTIONHEIGHT;
		}
		
		scrollh=(StyleFlags&S_FORCEHSCROLL)? 1:0;
		scrollv=(StyleFlags&S_FORCEVSCROLL)? 1:0;
		
		if(clientwidth>w) scrollh=1;
		if(clientheight>h) scrollv=1;

		if(scrollh) {
			h-=SCROLLBARWIDTH;
			if(!scrollv) {
				if(clientheight>h) scrollv=1;
			}
		}
		if(scrollv) {
			w-=SCROLLBARWIDTH;
			if(!scrollh) {
				if(clientwidth>w) {
					h-=SCROLLBARWIDTH;
					scrollh=1;
				}
			}
		}
		
		vieww=w;
		viewh=h;
		viewx=viewy=0;
		
		if(!Resize(clientwidth,clientheight)) InsufMemoryError();

		ClearClipRegion();

	
}

gForm::~gForm()
{
	
}

void gForm::Update(gUpdate *u)
{
	if(!(StateFlags&STAT_VISIBLE)) return;
	
	RedrawNonClient();
	if(SoftMenu) {
		if((StateFlags&STAT_FOCUS)) SoftMenu->Update(u);
	}

	gUpdate uclient;
	
	uclient.clipx=u->clipx-posx+viewx;
	uclient.clipy=u->clipy-posy+viewy;
	uclient.clipx2=u->clipx2-posx+viewx;
	uclient.clipy2=u->clipy2-posy+viewy;
	
	// CLEAR BACKGROUND IN CLIENT AREA
	
	

	// PASS UPDATE MESSAGE TO CHILDS
	
	gControl *chld=FirstChild;
	
	while(chld) {
		chld->Update(&uclient);
		chld=chld->Next;
	}
	
	// UPDATE VIEWED PORTION
	
	drawsurf.x=(u->clipx<posx)? viewx:(u->clipx-posx+viewx);
	drawsurf.y=(u->clipy<posy)? viewy:(u->clipy-posy+viewy);
	
	if(drawsurf.x>=viewx+vieww) return;
	if(drawsurf.y>=viewy+viewh) return;
	
	int dw,dh;
	dw=u->clipx2-posx+viewx-drawsurf.x+1;
	if(dw<=0) return;
	if(drawsurf.x+dw>=viewx+vieww) dw=viewx+vieww-drawsurf.x;
	
	dh=u->clipy2-posy+viewy-drawsurf.y+1;
	if(dh<=0) return;
	if(drawsurf.y+dh>=viewy+viewh) dh=viewy+viewh-drawsurf.y;

	
	Parent->drawsurf.x=drawsurf.x-viewx+posx;
	Parent->drawsurf.y=drawsurf.y-viewy+posy;
	
	ggl_bitblt(&(Parent->drawsurf),&drawsurf,dw,dh);
	

}

void gForm::Initialize(unsigned int *Data)
{

}
void gForm::Destroy(unsigned int *Data)
{
	
}

BOOL gForm::DoEvent(gEvent *e)
{
	

		if(e->Type==GEVENT_KEYB) {
			// SEND KEYBOARD ONLY TO FOCUSED CONTROL
			if(Focus) {
				if(Focus->DoEvent(e)) return TRUE;
			}
			if(DoEventExclusive(e)) return TRUE;
			
		}
		else {
			if(DoEventExclusive(e)) return TRUE;
	
gControl *ptr=FirstChild;


while(ptr) {
	if(ptr->DoEvent(e)) return TRUE;
	ptr=ptr->Next;
	}

		}
if(SoftMenu) SoftMenu->DoEvent(e);

return FALSE;	
}




BOOL gForm::DefaultEvent(gEvent *e)
{
	// TODO: IMPLEMENT FULL DEFAULT BEHAVIOR
	if(e->Type==GEVENT_GUI)
	{
		gUpdate u;
		
		switch(e->Arg)
		{
		case GEVGUI_FORMOPEN:
			StateFlags|=STAT_VISIBLE;
			if(!Focus)	{
				gControl *pr=FirstChild;
				while(pr) {
					if(pr->AcceptFocus()) {
						Focus=pr;
						break;
					}
					pr=pr->Next;
				}
			}
			u.clipx=ncx;
			u.clipy=ncy;
			u.clipx2=ncx+ncwidth;
			u.clipy2=ncy+ncheight;
			Parent->Invalidate(&u);
			return FALSE;

		case GEVGUI_FORMCLOSE:
			StateFlags&=~STAT_VISIBLE;
			u.clipx=ncx;
			u.clipy=ncy;
			u.clipx2=ncx+ncwidth;
			u.clipy2=ncy+ncheight;
			Parent->Invalidate(&u);
			return FALSE;
		
		case GEVGUI_LOSEFOCUS:
			StateFlags&=~STAT_FOCUS;
			if(Focus) Focus->DoEvent(e);
			HideSoftMenu();
			return TRUE;
			
		case GEVGUI_GETFOCUS:
			StateFlags|=STAT_FOCUS;
			if(Focus) Focus->DoEvent(e);
			ShowSoftMenu();
			return TRUE;
			
			
		// TODO: RESPOND TO EVENTS
			
		default:
			return FALSE;
		}
		
	}
	
	
	if(e->Type==GEVENT_KEYB) {
		// IMPLEMENT BASIC KEYBOARD HANDLING
	switch(e->Arg)
	{
	case KM_PRESS | KB_ADD | SHIFT_ALPHAHOLD:	
	++ GetApplication()->contrastlevel;
	GetApplication()->contrastlevel&=0xf;
	lcd_setcontrast(GetApplication()->contrastlevel);
	return TRUE;
	case KM_PRESS | KB_SUB | SHIFT_ALPHAHOLD:
	-- GetApplication()->contrastlevel;
	GetApplication()->contrastlevel&=0xf;
	lcd_setcontrast(GetApplication()->contrastlevel);
	return TRUE;
	case KM_PRESS | KB_UP:
	case KM_PRESS | KB_UP | SHIFT_RS | SHIFT_RSHOLD:
	case KM_PRESS | KB_UP | SHIFT_ALPHAHOLD:
	
	{
		gControl *pr,*start;
		if(Focus) start=Focus->Prev;
		else start=LastChild;
		if(!start) start=LastChild;
		
		pr=start;
		
		while(pr) {
			if(pr->AcceptFocus() && (pr!=Focus)) {
				gEvent ev;
				ev.Type=GEVENT_GUI;
				ev.Data=NULL;
				ev.Arg=GEVGUI_LOSEFOCUS;
				Focus->DoEvent(&ev);
				ev.Arg=GEVGUI_GETFOCUS;
				Focus=pr;
				Focus->DoEvent(&ev);
				return TRUE;
			}
			if(pr==FirstChild) pr=LastChild;
			else pr=pr->Prev;
			if(pr==start) break;
		}
		// NO CHILDS TO SWITCH FOCUS, DO SCROLL INSTEAD
			SetViewY(viewy-4);
			return TRUE;

	}
	break;
	case KM_PRESS | KB_DN :
	case KM_PRESS | KB_DN | SHIFT_RS | SHIFT_RSHOLD:
	case KM_PRESS | KB_DN | SHIFT_ALPHAHOLD:
	case KM_PRESS | KB_ENT:
	
	{
		gControl *pr,*start;
		if(Focus) start=Focus->Next;
		else start=FirstChild;
		if(!start) start=FirstChild;
		
		pr=start;
		
		while(pr) {
			if(pr->AcceptFocus() && (pr!=Focus)) {
				gEvent ev;
				ev.Type=GEVENT_GUI;
				ev.Data=NULL;
				ev.Arg=GEVGUI_LOSEFOCUS;
				if(Focus) Focus->DoEvent(&ev);
				ev.Arg=GEVGUI_GETFOCUS;
				Focus=pr;
				Focus->DoEvent(&ev);
				return TRUE;
			}
			if(pr==LastChild) pr=FirstChild;
			else pr=pr->Next;
			if(pr==start) break;
		}
		// NO CHILDS TO SWITCH FOCUS, DO SCROLL INSTEAD
		// DO NOT SCROLL USING ENTER KEY
		if(e->Arg!=(KM_PRESS | KB_ENT)) {
		SetViewY(viewy+4);
		return TRUE;
		}
		return FALSE;
	}
	break;
	case KM_PRESS | KB_UP | SHIFT_LS | SHIFT_LSHOLD:

		SetViewY(viewy-4);
		return TRUE;
	case KM_PRESS | KB_DN | SHIFT_LS | SHIFT_LSHOLD:
	SetViewY(viewy+4);
	return TRUE;
	case KM_PRESS | KB_LF :
	case KM_PRESS | KB_LF | SHIFT_LS | SHIFT_LSHOLD:
	SetViewX(viewx-4);
	return TRUE;
	case KM_PRESS | KB_RT:
	case KM_PRESS | KB_RT | SHIFT_LS | SHIFT_LSHOLD:
	SetViewX(viewx+4);
	return TRUE;

	default: 
		break;

	}
		
	}
	
	
	return FALSE;
	}

void gForm::RedrawBackground(gUpdate *u)
{
	int x1,y1,x2,y2;
	x1=clipx;
	x2=clipx2;
	y1=clipy;
	y2=clipy2;
	ClearClipRegion();		// THE CLIENT
	
	Rect(u->clipx,u->clipy,u->clipx2,u->clipy2,GWHITE);
	
	SetClipRegion(x1,y1,x2,y2);
}

void gForm::RedrawNonClient()
{
// DRAW NON-CLIENT ITEMS DIRECTLY INTO PARENT SURFACE
	gControl *P=Parent;
	int x,y,w,h;
	
	x=ncx;
	y=ncy;
	w=ncwidth;
	h=ncheight;
	if(StyleFlags&S_FORMBORDER) {

		P->HLine(y,x+1,x+w-2,GBLACK);
		P->HLine(y+h-1,x+1,x+w-2,GBLACK);
		P->VLine(x,y+1,y+h-2,GBLACK);
		P->VLine(x+w-1,y+1,y+h-2,GBLACK);
		
		x+=BORDERWIDTH;
		y+=BORDERWIDTH;
		w-=2*BORDERWIDTH;
		h-=2*BORDERWIDTH;
		
	}
	
	if(StyleFlags&S_TOPCAPTION ) {
		P->HLine(y,x,x+w-1,GGRAY3);
		P->HLine(y+1,x,x+w-1,GGRAY1);
		P->Rect(x,y+2,x+w-1,y+CAPTIONHEIGHT-5,GWHITE);
		P->HLine(y+CAPTIONHEIGHT-1,x,x+w-1,GGRAY7);
		P->HLine(y+CAPTIONHEIGHT-2,x,x+w-1,GGRAY5);
		P->HLine(y+CAPTIONHEIGHT-3,x,x+w-1,GGRAY3);
		P->HLine(y+CAPTIONHEIGHT-4,x,x+w-1,GGRAY1);

		gFont f(CAPTIONHEIGHT-2);
		
		int capx,capy;
		capx=x+w/2-f.TextWidth(Caption)/2;
		capy=y+CAPTIONHEIGHT/2-f.TextHeight()/2;
		// SHOW TITLE
		P->DrawText(capx,capy,Caption,&f,GBLACK);
		
	}
	// DRAW SCROLL BARS
	if(scrollh) DrawHScrollBar();
	if(scrollv) DrawVScrollBar();
	
	
}

void gForm::AddSoftMenu(gSoftMenu *menu)
{
	SoftMenu=menu;
	StyleFlags=(gFormStyles) (StyleFlags&(~(S_SINGLESOFTMENU|S_EXSOFTMENU)));
	StyleFlags=(gFormStyles) (StyleFlags|((menu->Style)&(S_SINGLESOFTMENU|S_EXSOFTMENU)));
	if(StyleFlags&S_FULLSCREEN) {
		if(StyleFlags & (S_SINGLESOFTMENU)) viewh-=SOFTMENUHEIGHT;
		else if(StyleFlags & (S_EXSOFTMENU)) viewh-=3*SOFTMENUHEIGHT;
	}
	menu->Visible=TRUE;
	Parent->Invalidate();
}


void gForm::HideSoftMenu()
{
	if(!SoftMenu) return;
	if((StyleFlags&S_FULLSCREEN) && (SoftMenu->Visible)) {
		if(StyleFlags & (S_SINGLESOFTMENU)) viewh+=SOFTMENUHEIGHT;
		else if(StyleFlags & (S_EXSOFTMENU)) viewh+=3*SOFTMENUHEIGHT;
	}
	SoftMenu->Visible=FALSE;
	Parent->Invalidate();
}

void gForm::ShowSoftMenu()
{
	if(!SoftMenu) return;
	if((StyleFlags&S_FULLSCREEN) && !(SoftMenu->Visible)) {
		if(StyleFlags & (S_SINGLESOFTMENU)) viewh-=SOFTMENUHEIGHT;
		else if(StyleFlags & (S_EXSOFTMENU)) viewh-=3*SOFTMENUHEIGHT;
	}
	SoftMenu->Visible=TRUE;
	Parent->Invalidate();
	
}
