//& *** (c) 2006-2011 The HPGCC3 Team ***
//& Claudio Lapilli
//& Ingo Blank
//&
//& This file is licensed under the terms and conditions of the
//& HPGCC3 license that is included with the source distribution.
//& *** (c) 2006-2011 The HPGCC3 Team ***

#include <hpgcc3.h>
#include <master_globals.h>

extern unsigned int __startup_options;

void __tmr_eventreschedule();


#define HWREG(base,off) ( (volatile unsigned int *) (((int)base+(int)off)))


// SAVE TIMERS CONFIGURATION - 13 WORDS REQUIRED
void tmr_save(unsigned int *tmrbuffer)
{

memcpyw(tmrbuffer,(void *)HWREG(TMR_REGS,0),5);

tmrbuffer[5]=*HWREG(TMR_REGS,0x18);
tmrbuffer[6]=*HWREG(TMR_REGS,0x1c);
tmrbuffer[7]=*HWREG(TMR_REGS,0x24);
tmrbuffer[8]=*HWREG(TMR_REGS,0x28);
tmrbuffer[9]=*HWREG(TMR_REGS,0x30);
tmrbuffer[10]=*HWREG(TMR_REGS,0x34);
tmrbuffer[11]=*HWREG(TMR_REGS,0x3c);
tmrbuffer[12]=*HWREG(INT_REGS,0x8);	// GET TIMERS INTERRUPT MASK
}


// RESTORE TIMERS

void tmr_restore(unsigned int *tmrbuffer)
{
// MASK ALL TIMER INTERRUPTS
*HWREG(INT_REGS,0x8)|=0x7c00;

// STOP ALL RUNNING TIMERS
*HWREG(TMR_REGS,8)&=0xFF8000e0;

// CLEAN ALL PENDING INTERRUPTS
*HWREG(INT_REGS,0x10)=0x7c00;
*HWREG(INT_REGS,0)=0x7c00;



// RESTORE REGISTERS
*HWREG(TMR_REGS,0)=tmrbuffer[0];
*HWREG(TMR_REGS,4)=tmrbuffer[1];
*HWREG(TMR_REGS,0xc)=tmrbuffer[3];
*HWREG(TMR_REGS,0x10)=tmrbuffer[4];

*HWREG(TMR_REGS,0x18)=tmrbuffer[5];
*HWREG(TMR_REGS,0x1c)=tmrbuffer[6];
*HWREG(TMR_REGS,0x24)=tmrbuffer[7];
*HWREG(TMR_REGS,0x28)=tmrbuffer[8];
*HWREG(TMR_REGS,0x30)=tmrbuffer[9];
*HWREG(TMR_REGS,0x34)=tmrbuffer[10];
*HWREG(TMR_REGS,0x3c)=tmrbuffer[11];

*HWREG(INT_REGS,0x8)=(*HWREG(INT_REGS,0x8)&(~0x7c00))|(tmrbuffer[12]&0x7c00);

// FINALLY RESTART ALL TIMERS
*HWREG(TMR_REGS,8)=tmrbuffer[2];
}


void __tmr_irqservice()
{
__mg->__systmr+=0x10000;

}

void __tmr_irqeventsvc()
{

__mg->__evtmr+=*HWREG(TMR_REGS,0x1c)+1;	// TCMP1 HOLDS PREVIOUS RELOAD VALUE
*HWREG(TMR_REGS,0x1c)=*HWREG(TMR_REGS,0x18);	// PRESERVE CURRENT LOAD VALUE

int nextticks=0x7fffffff;
register int ntck,active;
int f;

for(active=0,f=0;f<NUM_EVENTS;++f)
{
if(__mg->tmr_events[f].status&1) {

		++active;
	// EVENT IS ACTIVE, CHECK IF TIME TO ACTIVATE
	if((__mg->tmr_events[f].ticks==__mg->__evtmr) || (__mg->tmr_events[f].status&4)) 
	{
		// EXECUTE EVENT
		(*(__mg->tmr_events[f].eventhandler))();
		if((__mg->tmr_events[f].status&3)==3) {
			// AUTORELOAD
			__mg->tmr_events[f].ticks+=__mg->tmr_events[f].delay;
			__mg->tmr_events[f].status&=~4;
		}
		else { __mg->tmr_events[f].status&=~5; continue; }
	}
	
	// CALCULATE NEXT TIMER DELAY
	ntck=__mg->tmr_events[f].ticks-__mg->__evtmr-*HWREG(TMR_REGS,0x1c)-1;
	
	// IF ntck<0 THE EVENT WILL BE MISSED, EXECUTED LATE ON NEXT INTERRUPT
	// IF ntck==0 THE EVENT WILL HAPPEN NEXT TIMER INTERRUPT

	if(ntck<=0) {
	if(ntck<0) {
		// THIS CAN ONLY HAPPEN WHEN A NEW EVENT IS ADDED/REMOVED,
		// AND ONLY ON RARE CASES
		// THE EVENT WILL BE EXECUTED LATE (FEW usec) ON THE NEXT INTERRUPT
		__mg->tmr_events[f].status|=4; // REQUEST FORCED EVENT EXECUTION ON NEXT INTERRUPT
	}
	
	// CATCH UP WITH COUNTERS IF MULTIPLE EVENTS MISSED
	while(ntck<=0) {
		ntck+=__mg->tmr_events[f].delay;
	__mg->tmr_events[f].ticks+=__mg->tmr_events[f].delay;
	}
	__mg->tmr_events[f].ticks-=__mg->tmr_events[f].delay;
	}
	if( (ntck<nextticks) && (ntck>=0)) nextticks=ntck;	// RECORD AS NEXT CLOCK	
		
	}
}

if(!active) {
// NO MORE EVENTS ARE ACTIVE, STOP TIMER UNTIL NEW EVENTS ARE SCHEDULED
*HWREG(TMR_REGS,0x8)&=~0xf00;
__mg->__evtmr=0;
return;
}

if(nextticks>0x10800) nextticks=0x10000;
if(nextticks>0x10000) nextticks=0x8000;

*HWREG(TMR_REGS,0x18)=nextticks-1;

}





// SYSTEM TIMER FREQUENCY = 100 KHz
// ALTERNATIVELY, SWITCH TO 125 KHz IF 100 KHz CANNOT BE CORRECTLY REPRESENTED
#define SYSTIMER_FREQ 100000
#define ALTSYSTIMER_FREQ 125000


// PREPARE TIMERS FOR HPGCC FUNCTIONS	
void tmr_setup()
{

// MASK ALL TIMER INTERRUPTS
*HWREG(INT_REGS,0x8)|=0x7c00;

// STOP ALL RUNNING TIMERS
*HWREG(TMR_REGS,8)&=0xFF8000e0;

// CLEAN ALL PENDING INTERRUPTS
*HWREG(INT_REGS,0x10)=0x7c00;
*HWREG(INT_REGS,0)=0x7c00;


// START TIMER0 AS 64-BIT RUNNING TIMER AT 100 KHz

__mg->__systmr=0;

unsigned int pclk=__cpu_getPCLK();

unsigned int divider,prescaler;

prescaler=(pclk<<3)/SYSTIMER_FREQ;
divider=1;

if(prescaler&0xf) {
// WARNING: REQUESTED FREQUENCY CANNOT BE REPRESENTED ACCURATELY
// THIS ONLY HAPPENS AT 75 MHz, ALL OTHER SPEEDS ARE OK
prescaler=(pclk<<3)/ALTSYSTIMER_FREQ;
}

while( prescaler> (1<<(11+divider)) ) { divider++; }

prescaler+= (1<<(2+divider));
prescaler>>=divider+3;

//if(divider>4) PCLK TOO HIGH TO ACHIEVE TIMER FREQUENCY, USE HIGHER MULTIPLE
if(divider>4) divider=4;

// CALCULATE SYSTEM CLOCK FREQUENCY
__mg->__sysfreq=(((pclk<<3)/prescaler) +  (1<<(divider+2)) )>>(divider+3);

// SET PRESCALER VALUES FOR TIMERS 0 AND 1
*HWREG(TMR_REGS,0)=(*HWREG(TMR_REGS,0)&(~0xFF)) | (prescaler-1);
*HWREG(TMR_REGS,4)=(*HWREG(TMR_REGS,4)&(~0xf000ff)) | (divider-1) | ((divider-1)<<4);

// SET COUNT VALUES TO MAXIMUM
*HWREG(TMR_REGS,0xc)=0xffff;
*HWREG(TMR_REGS,0x10)=0;
*HWREG(TMR_REGS,0x18)=0xffff;
*HWREG(TMR_REGS,0x1C)=0;

// PREPARE TIMED EVENTS HANDLER
memset((void *)__mg->tmr_events,0,sizeof(timed_event)*NUM_EVENTS);
__mg->__evtmr=0;


__irq_addhook(10,(__interrupt__) &__tmr_irqservice);
__irq_addhook(11,(__interrupt__) &__tmr_irqeventsvc);

// UNMASK INTERRUPTS FOR TIMERS 0 AND 1
*HWREG(INT_REGS,0x8)&=~0xc00;



// SET MANUAL UPDATE BIT
*HWREG(TMR_REGS,8)|=2;
// START TIMER0 
*HWREG(TMR_REGS,8)=(*HWREG(TMR_REGS,8)&(~0x1f)) | 0x9;

}

// FIX TIMERS SPEED WHEN CPU CLOCK IS CHANGED, CALLED FROM cpu_setspeed	
void __tmr_fix()
{
	if((__startup_options&_NO_TIMERS)) return;
// MASK ALL TIMER INTERRUPTS
//*HWREG(INT_REGS,0x8)|=0x7c00;

// STOP ALL RUNNING TIMERS
//*HWREG(TMR_REGS,8)&=0xFF8000e0;

// CLEAN ALL PENDING INTERRUPTS
//*HWREG(INT_REGS,0x10)=0x7c00;
//*HWREG(INT_REGS,0)=0x7c00;


// START TIMER0 AS 64-BIT RUNNING TIMER AT 100 KHz


unsigned int pclk=__cpu_getPCLK();

unsigned int divider,prescaler;

prescaler=(pclk<<3)/SYSTIMER_FREQ;
divider=1;

if(prescaler&0xf) {
// WARNING: REQUESTED FREQUENCY CANNOT BE REPRESENTED ACCURATELY
// THIS ONLY HAPPENS AT 75 MHz, ALL OTHER SPEEDS ARE OK
prescaler=(pclk<<3)/ALTSYSTIMER_FREQ;
}

while( prescaler> (1<<(11+divider)) ) { divider++; }

prescaler+= (1<<(2+divider));
prescaler>>=divider+3;

//if(divider>4) PCLK TOO HIGH TO ACHIEVE TIMER FREQUENCY, USE HIGHER MULTIPLE
if(divider>4) divider=4;

// CALCULATE SYSTEM CLOCK FREQUENCY
__mg->__sysfreq=(((pclk<<3)/prescaler) +  (1<<(divider+2)) )>>(divider+3);

// SET PRESCALER VALUES FOR TIMERS 0 AND 1
*HWREG(TMR_REGS,0)=(*HWREG(TMR_REGS,0)&(~0xFF)) | (prescaler-1);
*HWREG(TMR_REGS,4)=(*HWREG(TMR_REGS,4)&(~0xf000ff)) | (divider-1) | ((divider-1)<<4);


}











tmr_t tmr_getsysfreq()
{
	return __mg->__sysfreq;
}


tmr_t tmr_ticks()
{
unsigned int *timer0=(unsigned int *)HWREG(TMR_REGS,0x14);

unsigned int before;
ULONGLONG ticks1,ticks2;

do {
ticks1=__mg->__systmr;
before=*timer0;
ticks2=__mg->__systmr;
} while(ticks1!=ticks2);

return ticks1+0x10000-before;	
}


// RETURN DELAY IN MILLISECONDS
int tmr_ticks2ms(tmr_t before, tmr_t after)
{
	return ((after-before)*1000)/__mg->__sysfreq;
}

// RETURN DELAY IN MICROSECONDS
int tmr_ticks2us(tmr_t before, tmr_t after)
{
	return ((after-before)*1000000)/__mg->__sysfreq;
}

// ADD/SUBTRACT AN INTERVAL IN MILLISECONDS TO THE GIVEN TIME IN TICKS
tmr_t tmr_addms(tmr_t time, int ms)
{
	return time+((ms*__mg->__sysfreq)/1000);
}

// ADD/SUBTRACT AN INTERVAL IN MICROSECONDS TO THE GIVEN TIME IN TICKS
tmr_t tmr_addus(tmr_t time, int us)
{
	return time+((us*__mg->__sysfreq)/1000000);
}


void tmr_delayms(int milliseconds)
{
	if(milliseconds<=0) return;
tmr_t start=tmr_ticks();

// CALCULATE ENDING TICKS
tmr_t end=start+((milliseconds*__mg->__sysfreq)/1000);

// AND WAIT
while(end>tmr_ticks());

}
	
void tmr_delayus(int microseconds)
{
	if(microseconds<=0) return;

	ULONGLONG start=tmr_ticks();

// CALCULATE ENDING TICKS
ULONGLONG end=start+((microseconds*tmr_getsysfreq())/1000000);

// AND WAIT
while(end>tmr_ticks());

}

// WAIT UNTIL THE TIME REACHES A CERTAIN VALUE
void tmr_waituntil(tmr_t time)
{
	while(tmr_ticks()<time);
	return;
}




// RETURN AN EVENT HANDLER
HEVENT tmr_eventcreate(__interrupt__ handler,unsigned int ms,int autorepeat)
{
int f;

// NOTE: EVENT 0 IS RESERVED FOR THE SYSTEM (KEYBOARD DRIVER)

for(f=1;f<NUM_EVENTS;++f)
{
	if(!(__mg->tmr_events[f].status&1)) {
	// FOUND AVAILABLE EVENT	
		__mg->tmr_events[f].eventhandler=handler;
		__mg->tmr_events[f].delay=(ms*tmr_getsysfreq())/1000;
		if( (*HWREG(TMR_REGS,0x8))&0x100) {
		__mg->tmr_events[f].ticks=__mg->__evtmr+__mg->tmr_events[f].delay+*HWREG(TMR_REGS,0x1c)-*HWREG(TMR_REGS,0x20);
		}
		else __mg->tmr_events[f].ticks=__mg->__evtmr+__mg->tmr_events[f].delay;
		
		__mg->tmr_events[f].status=((autorepeat)? 2:0)|1;

		__tmr_eventreschedule();
	return f;
	}
}
return -1;
}


void __tmr_eventreschedule()
{
register int before;

if((*HWREG(TMR_REGS,0x8)) &0x100) {
before=*HWREG(TMR_REGS,0x20);
if(before<50) {	// MORE THAN 1 ms TO NEXT IRQ
	*HWREG(TMR_REGS,0x18)=50;	// GENERATE NEXT TIMER IRQ IN 1 ms
	return;
}
} else {
*HWREG(TMR_REGS,0x1c)=0;		// SET PREVIOUS LAP WAS NULL 
before=0;	
}

*HWREG(TMR_REGS,0x8)&=~0xf00;		// STOP TIMER
	
__mg->__evtmr+=*HWREG(TMR_REGS,0x1c)+1-before;

*HWREG(TMR_REGS,0x18)=50;		// SET NEXT RELOAD IN 1 ms
*HWREG(TMR_REGS,0x1c)=0;		// SET PREVIOUS LAP WAS NULL 
*HWREG(TMR_REGS,0x8)|=0x200;		// MANUAL UPDATE

*HWREG(TMR_REGS,0x8)=(*HWREG(TMR_REGS,0x8)&(~0xf00))|0x900;		// RESTART TIMER
return;
}




void tmr_eventkill(HEVENT event)
{
	if(event<0 || event>=NUM_EVENTS) return;
	__mg->tmr_events[event].status=0;	// KILL EVENT, TIMER WILL STOP AUTOMATICALLY

}

