#!/usr/bin/env python

#//& *** (c) 2006-2011 The HPGCC3 Team ***
#//& Claudio Lapilli
#//& Ingo Blank
#//&
#//& This file is licensed under the terms and conditions of the
#//& HPGCC3 license that is included with the source distribution.
#//& *** (c) 2006-2011 The HPGCC3 Team ***



# $Header: sources/romlib/jumptable_new.py,r226 2006-03-20 19:51:36 ingo $

# --- jumptable.py
# --- Version 0.2 2006-03-20 ibl
# ---
# --- (I)
# --- Read an input file with symbols, one per line, and generate
# --- a C code file with a corresponding jumptable.
# --- Exclude symbols marked with an trailing 'x' or 'X'.
# ---
# --- (II)
# --- Invoke 'make' with a special Makefile on the C file.
# ---
# --- (III)
# --- Parse the symbol table of the generated exe and append all unresolved
# --- references to the original input file.
# ---
# --- (IV)
# --- Repeat the process recursively, until all symbols are resolved.


import os,re,sys
from sets import Set

try:
    import psyco
    psyco.full()
except:
    pass
    


def get_symbol_lists(name):
    
    f = open(name,"r")
    
    ordered_include = []
    exclude = Set()
    
    for zz in f:
        zz = zz.strip()
        if not zz:
            continue
        if zz[0] == '#':
            continue
        if len(zz) >= 2 and zz[:2] == '//':
            continue
        
        m = RE_exclusion_pattern.search(zz)
        if m:
            sym = m.group(1)
            if m.group(2) in ('x','X'):
                exclude.add(sym)
            else:
                if not sym in ordered_include:
                    ordered_include.append(sym)
                    
        else:
            if not zz in ordered_include:
                ordered_include.append(zz)
                
            
    
    f.close()
    
    return ordered_include,exclude



def generate_jump_table(name,ordered_syms):
    
    out = open(name,"w")
    n = 0
    txt = """/* --- AUTOMATICALLY GENERATED CODE, DO NOT EDIT! --- */

""" 

    out.write("%s\n" % txt)
    for f in ordered_syms:
        n += 1
        txt = """// Ordinal %d
void %s%s() __attribute__((naked));
void %s%s() { asm(\"b %s\"); }""" % (n,ROM_PREFIX,f,ROM_PREFIX,f,f)
        
        out.write("%s\n" % txt)
        
    out.close()
    
    
# --- START OF MAIN SCRIPT ---

# --- CONSTANTS (modify if appropriate)

ROM_PREFIX = '_ROM_'
MAKEFILE = 'RL_Makefile'
DEFAULT_OUTPUT = '__romlib.c'


# --- REGULAR EXPRESSIONS FOR INPUT PARSING

RE_dump_ROM_entry = re.compile(r'((\d|[a-f])+)\s+<' + ROM_PREFIX + r'([a-zA-z0-9_]+)>:')
RE_dump_lib_entry = re.compile(r'((\d|[a-f])+)\s+<([a-zA-Z_][a-zA-z0-9_]*)>:')
RE_switch_stub = re.compile(r'_(from|to)_(arm|thumb)')
RE_exclusion_pattern = re.compile(r'([a-zA-Z_][a-zA-z0-9_]*)\s+([a-zA-z])')

# --- VARS

rom_entries = Set()
lib_entries = Set()

try:
    file_name = sys.argv[1]
except:
    print "Use: %s <organized_list.list> [<%s>]" % (os.path.split(sys.argv[0])[1],DEFAULT_OUTPUT)
    sys.exit(1)
    
ordered_gensyms,exclude = get_symbol_lists(file_name)
    
try:
    out_name = sys.argv[2]
except:
    out_name = DEFAULT_OUTPUT

generate_jump_table(out_name,ordered_gensyms)

# Now make the ROM executable ...

rom_base,ext = os.path.splitext(out_name)
rom_exe = rom_base + ".exe"
cmd = "make -f %s %s" % (MAKEFILE,rom_exe)
res = os.system(cmd)

# res is system dependent, check if appropriate ...

if res != 0:
    print "Make Error, aborting."
    sys.exit(1)


# ... and read the symbol table piped in from 'objdump' ...

cmd = "arm-none-eabi-objdump -D -marm %s" % rom_exe

inp = os.popen(cmd)

for zz in inp:
    
    if zz.find('Disassembly of section .rodata:') >= 0:
        break
    
    m = RE_dump_ROM_entry.search(zz)
    if m:
        rom_entries.add(m.group(3))
        continue
    
    m = RE_dump_lib_entry.search(zz)
    if m:
        if not RE_switch_stub.search(zz):
            lib_entries.add(m.group(3))
        continue
    
inp.close()

table = (lib_entries.difference(rom_entries)).difference(exclude)

njumps = len(table)

# we append the symbols gathered in 'table' to our original symbol set

if njumps:
    
    # Make a backup copy of the input list...
    new_name = file_name + '~'
    cmd = "cp %s %s" % (file_name,new_name)
    os.system(cmd)
    
    out = open(file_name,"a")
    
    txt = "// --- BEGIN OF AUTO APPENDED SYMBOLS ---"
    out.write("\n%s\n\n" % txt)
    
    for f in table:
        out.write("%s\n" % f)
        
    txt = "// --- END OF AUTO APPENDED SYMBOLS ---"
    out.write("\n%s\n\n" % txt)
    
    out.close()
    
    
    
    # ... finally call ourselves recursively ...
    
    cmd = "%s %s" % (sys.argv[0],file_name)
    os.system(cmd)
    
    # Done !
    
